/*
 * Copyright Bigeon Emmanuel (2014)
 *
 * emmanuel@bigeon.fr
 *
 * This software is a computer program whose purpose is to
 * provide a generic framework for console applications.
 *
 * This software is governed by the CeCILL license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * As a counterpart to the access to the source code and  rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty  and the software's author,  the holder of the
 * economic rights,  and the successive licensors  have only  limited
 * liability.
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms.
 */
/**
 * gclc:net.bigeon.gclc.system.SystemConsoleManager.java
 * Created on: Dec 19, 2014
 */
package net.bigeon.gclc.utils;

/*-
 * #%L
 * Generic Command Ligne console
 * %%
 * Copyright (C) 2014 - 2018 bigeon.fr
 * %%
 * This software is governed by the CeCILL  license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/ or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * As a counterpart to the access to the source code and  rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty  and the software's author,  the holder of the
 * economic rights,  and the successive licensors  have only  limited
 * liability.
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms.
 * #L%
 */
import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.PrintStream;
import java.nio.charset.Charset;

import net.bigeon.gclc.manager.ConsoleInput;
import net.bigeon.gclc.tools.ConstantString;
import net.bigeon.gclc.tools.StringProvider;

/** A console using the input stream and print stream.
 * <p>
 * The default constructor will use the system standart input and output.
 *
 * @author Emmanuel BIGEON */
public final class StreamConsoleInput implements ConsoleInput {

    /** The default prompt. */
    public static final StringProvider DEFAULT_PROMPT = new ConstantString("> "); //$NON-NLS-1$

    /** The command prompt. It can be changed. */
    private StringProvider prompt = DEFAULT_PROMPT;

    /** The print stream. */
    private final PrintStream    out;
    /** The input stream. */
    private final BufferedReader in;

    /** If the manager is closed. */
    private boolean closed = false;

    /** The prompting thread. */
    private final Thread promptThread;

    /** The reading runnable. */
    private final ReadingRunnable reading;

    /** This default constructor relies on the system defined standart output and
     * input stream. */
    public StreamConsoleInput() {
        this(System.out, System.in, Charset.defaultCharset());
    }

    /** Create the stream base console input.
     *
     * @param out the output stream
     * @param in the input stream
     * @param charset the charset for the input */
    public StreamConsoleInput(final PrintStream out, final InputStream in,
                              final Charset charset) {
        super();
        this.out = out;
        this.in = new BufferedReader(new InputStreamReader(in, charset));
        reading = new ReadingRunnable(this.in);
        promptThread = new Thread(reading, "prompt"); //$NON-NLS-1$
        promptThread.setDaemon(true);
        promptThread.start();
    }

    /** Check that the console input is not closed.
     *
     * @throws IOException if the stream was closed */
    private void checkOpen() throws IOException {
        if (closed) {
            throw new IOException();
        }
    }

    /* (non-Javadoc)
     * @see net.bigeon.gclc.manager.ConsoleManager#close() */
    @Override
    public void close() {
        closed = true;
        reading.setRunning(false);
        promptThread.interrupt();
    }

    /* (non-Javadoc)
     * @see net.bigeon.gclc.manager.ConsoleInput#getPrompt() */
    @Override
    public StringProvider getPrompt() {
        return prompt;
    }

    /** Beware, in this implementation this is the same as closing the manager.
     *
     * @see net.bigeon.gclc.manager.ConsoleInput#interruptPrompt() */
    @Override
    public void interruptPrompt() {
        reading.interrupt();
    }

    /* (non-Javadoc)
     * @see net.bigeon.gclc.manager.ConsoleManager#isClosed() */
    @Override
    public boolean isClosed() {
        return closed;
    }

    /* (non-Javadoc)
     * @see net.bigeon.gclc.ConsoleManager#prompt() */
    @Override
    public String prompt() throws IOException {
        return prompt(prompt.apply());
    }

    /* (non-Javadoc)
     * @see net.bigeon.gclc.manager.ConsoleManager#prompt(long) */
    @Override
    public String prompt(final long timeout) throws IOException {
        return prompt(prompt.apply(), timeout);
    }

    /* (non-Javadoc)
     * @see net.bigeon.gclc.ConsoleManager#prompt(java.lang.String) */
    @Override
    public String prompt(final String message) throws IOException {
        checkOpen();
        if (out != null) {
            out.print(message);
            out.flush();
        }
        return reading.getMessage();
    }

    /* (non-Javadoc)
     * @see net.bigeon.gclc.ConsoleManager#prompt(java.lang.String) */
    @Override
    public String prompt(final String message, final long timeout) throws IOException {
        checkOpen();
        if (out != null) {
            out.print(message);
            out.flush();
        }
        return reading.getNextMessage(timeout);
    }

    /* (non-Javadoc)
     * @see net.bigeon.gclc.manager.ConsoleInput#setPrompt(java.lang.String) */
    @Override
    public void setPrompt(final String prompt) {
        this.prompt = new ConstantString(prompt);
    }

    @Override
    public void setPrompt(StringProvider string) {
        prompt = string;
    }
}
